'''
Saravanan Ramanathan
Updated: November 2016
'''

# MCFairGen: Generate multi-core dual-criticality task sets

from taskparam import ts
import mrand as MR
from math import *
from random import *

import pylab
import random
import numpy


# Function to generate all task sets
def run():									# Generate tasksets
    for m in ts.m:								# no. of cores
	for hhu in range(ts.hhb_l,ts.hhb_h+1,ts.hhb_step):			# HI-Utilization of HC-tasks
		hhubound=m*float(hhu)/100					# Total bound = m * hhu
		for hlu in range(ts.hlb_l,hhu,ts.hlb_step):			# LO-Utilization of HC-tasks
			hlubound=m*float(hlu)/100				# Total bound = m * hlu
			for lu in range(ts.lb_l,100-hlu+1,ts.lb_step):		# LO-Utilization of LC-tasks
				lubound=m*float(lu)/100				# Total bound = m * lu
                    		for p in ts.t_prob:				# Percentage of HI-tasks
				   	NHmin = ceil(hhubound/ts.u_h)		# Calculate minimum required number of HC-tasks
				    	NLmin = ceil(lubound/ts.u_h)		# Calculate minimum required number of LC-tasks
                        		loop(m,hhu,hlu,lu,p,hhubound,hlubound,lubound,NHmin,NLmin)# Generate now

# Generate task set with given parameters                                
def loop(m,hhu,hlu,lu,p,hhubound,hlubound,lubound,NHmin,NLmin):

    print m,hhu,hlu,lu,p							# Print the parameters of a task set
    prefix=str(m)+"_"+str(hhu)+"_"+str(hlu)+"_"+str(lu)+"_"+str(p)		# Prefix to write down task sets to a file
    f = open(ts.path+"_"+prefix+".in","w")					# Open a file in write mode to write the generated task sets

    cond =  1									# Loop condition variable

    for j in range(ts.numTask):							# For total number of task sets
	ai = max(m+1,ceil(NHmin/p),ceil(NLmin/(1-p)))				# Minimum required total number of tasks in a task set
	if(ai >= 10*m):								# Total tasks upper bounded by 10*m
		Ntotal = 10*m
	else:									# Vary the total number of tasks for each task set
		Ntotal = randint (ai, 10*m)

	NH = int(max((p*Ntotal),NHmin))						# Calculate total number of HC-tasks
	NL = int(Ntotal-NH)							# Calculate total number of LC-tasks

        while cond:
            tasks=generate(m,hhubound,hlubound,lubound,p,NH,NL)			# Call task generate function	
	    if tasks!=None:							# If task set not empty
		    writeTasks(f,tasks)						# Write task set into file
		    break
        f.write("---%d\n"%(j))							# Write taskset count
    f.close()									# Close file

# Function to generate task
def generate(m,hhubound,hlubound,lubound,p,NH,NL):

    Ull = 0									# LC - LO-Utilization vector
    Uhh = 0									# HC - HI-Utilization vector
    Uhl = 0									# HC - LO-Utilization vector

    num = 0									# Task count
    temp_task = []								# Temporary task set

    # HC tasks
    qh,q1= MR.randfixedsum(NH,1,hhubound,ts.u_l,ts.u_h)				# Use mRandFixedSum to calculate u_i^H of HC-tasks
    ql,q1 = pick_low_util(qh,NH,hlubound,hhubound,ts.u_l)			# Use BoundedUniform to calculate u_i^L of HC-tasks

    for j in range(NH):								# For all HC-tasks

	if(ts.t_log_u==0):
		T = int(uniform(ts.t_l,ts.t_h))					# Pick uniform period in range (pl, ph)
	else:
		T = int(floor((exp(uniform(log(ts.t_l),log(ts.t_h+ts.t_l))))/ts.t_l)*ts.t_l)	# Pick log-uniform period in range (pl, ph)

	uh = qh[j]								# Pick Hi utilization from list
        ul = ql[j]			        				# Pick Lo utilization from list

	crit = 1								# Define criticality 1-HC 0-LC

        c_lo = ul*T								# Compute execution time (non-integer)
        c_hi = uh*T

        if (c_lo > c_hi+0.000001):						# Return error if CiL > CiH
		print c_lo,c_hi
		print "CError: Execution contraint not met"
		c = raw_input("Enter")

	Uhl+=float(c_lo)/T							# Add utilization values to corresponding vectors
	Uhh+=float(c_hi)/T

	if (ceil(c_hi) < T):							# Compute deadline of a task
	    if(ts.d_log_u==0):
		d=randint(ceil(c_hi),T)						# Compute uniform deadline for constrained deadline task systems
	    else:
		d=int(floor((exp(uniform(log(ceil(c_hi)),log(T))))/ceil(c_hi))*ceil(c_hi))	# Compute log-uniform deadline for constrained deadline task systems
	else:
		d=T

	if ((Uhh>hhubound+0.000001) or (Uhl>hlubound+0.000001)):		# If total utilization greater than bound then break
		print "BError: Utilization bound exceeded"
		c = raw_input("Enter")

	if(c_hi<=0 or c_lo<=0):							# Append task if execut. requirement is > 0
		print "CError: Invalid execution time"
		c = raw_input("Enter")
	
	temp_task.append((T,c_lo,c_hi,crit,d))					# Append period, exec. time, criticality, deadline

        num = num + 1								# Increase count - no. of tasks

    # LC Tasks
    x,xl= MR.randfixedsum(NL,1,lubound,ts.u_l,ts.u_h)  				# Use mRandFixedSum to calculate u_i^L of LC-tasks

    for j in range(NL):								# For each LC-task
		if(ts.t_log_u==0):
			T = int(uniform(ts.t_l,ts.t_h))				# Pick uniform period in range (pl, ph)
		else:
			T = int(floor((exp(uniform(log(ts.t_l),log(ts.t_h+ts.t_l))))/ts.t_l)*ts.t_l)	# Pick log-uniform period in range (pl, ph)

		ul = x[j]							# Pick LO-utilization from list
		uh = ul								# Assign HI-utilization = LO-utilization

		crit = 0							# Define criticality 1-HC 0-LC

        	c_lo = ul*T							# Compute execution time (non-integer)
        	c_hi = uh*T

		Ull+=float(c_lo)/T						# Add utilization values to corresponding vector

		if (ceil(c_hi) < T):
			if(ts.d_log_u==0):
				d=randint(ceil(c_hi),T)				# Compute uniform deadline for constrained deadline task systems
	    		else:
				d=int(floor((exp(uniform(log(ceil(c_hi)),log(T))))/ceil(c_hi))*ceil(c_hi))	# Compute log-uniform deadline for constrained deadline task systems
		else:
			d=T

        	if (Ull>lubound+0.000001):					# If total utilization greater than bound then break
			print "BError: Utilization bound exceeded"
			c = raw_input("Enter")

		if(c_lo<=0):							# Append task if execut. requirement is > 0
			print "CError: Invalid execution time"
			c = raw_input("Enter")

		temp_task.append((T,c_lo,c_hi,crit,d))				# Append period, exec. time, criticality, deadline

        	num = num + 1							# Increase count - no. of tasks

    if ((Uhh<hhubound-m*float(ts.hhb_step)/100) or (Uhl<hlubound-m*float(ts.hlb_step)/100) or (Ull<lubound-m*float(ts.lb_step)/100)):	# Utilization bound error
        return None								# Return none
    
    return temp_task								# Return generated task set


# Use BoundedUniform to calculate u_i^L of HC-tasks
def pick_low_util(qh,NH,hlubound,hhubound,u_l):
	
	B = hlubound								# Initialize total LO-Utilization Bound of HC-tasks
	Basign = 0
	Nrem = NH-1								# Initialize remaining number of HC-tasks to determine u_i^L
	Hrem = hhubound								# Initialize remaining HI-Utilization Bound of HC-tasks
	lh=list(numpy.array(qh).reshape(-1,))					# Computed u_i^H vector
	ql=lh									# Temporary u_i^L vector
	for i in range(NH):							# For each HC-task
		Hrem -= lh[i]							# Reduce u_i^H from remaining HI-Utilization Bound
		ql[i] = uniform(max(u_l,B-Hrem),min((B-(Nrem*u_l)),lh[i]))	# Compute u_i^L such that u_i^L<=u_i^H and \sum(u_i^L)<=bound
		Nrem -= 1							# Update count - remaining HC-tasks
		B -= ql[i]							# Update remaining bound
	#ql[NH-1] = B

	for i in range(NH):							# Compute total utilization bound for assigned u_i^L
		Basign += ql[i]

	if(Basign > hlubound+0.000001):						# Check if the total bound is acceptable
		c = raw_input("Enter +")
	if(Basign < hlubound-0.000001 ):
		c = raw_input("Enter -")

	return ql,1								# Return the generated u_i^L list

# Function to write a task set into file
def writeTasks(f,tasks):
    for task in tasks:
        f.write("%d %f %f %d %d\n"%(task[0],task[1],task[2],task[3],task[4]))   # Ordering of writing


def main():									# Main function
	run()

main()										# Call main()
